@@ -13,8 +13,8 @@ np = ak.nplike.NumpyMetadata.instance()
 
 def code_to_function(code, function_name, externals=None, debug=False):
     if debug:
-        print("################### " + function_name)  # noqa: T001
-        print(code)  # noqa: T001
+        print("################### " + function_name)  # noqa: T201
+        print(code)  # noqa: T201
     namespace = {} if externals is None else dict(externals)
     exec(code, namespace)
     return namespace[function_name]
@@ -105,7 +105,7 @@ class ArrayView:
     @classmethod
     def fromarray(cls, array):
         behavior = ak._v2._util.behavior_of(array)
-        layout = ak._v2.operations.convert.to_layout(
+        layout = ak._v2.operations.to_layout(
             array,
             allow_record=False,
             allow_other=False,
@@ -496,7 +496,7 @@ class RecordView:
     @classmethod
     def fromrecord(cls, record):
         behavior = ak._v2._util.behavior_of(record)
-        layout = ak._v2.operations.convert.to_layout(
+        layout = ak._v2.operations.to_layout(
             record,
             allow_record=True,
             allow_other=False,
@@ -663,12 +663,12 @@ class type_getattr_record(numba.core.typing.templates.AttributeTemplate):
 
                     def generic(self, args, kwargs):
                         if len(kwargs) == 0:
-                            sig = typer(recordviewtype, args)
+                            sig = typer(recordviewtype, args)  # noqa: B023
                             sig = numba.core.typing.templates.Signature(
                                 sig.return_type, sig.args, recordviewtype
                             )
                             numba.extending.lower_builtin(
-                                methodname,
+                                methodname,  # noqa: B023
                                 recordviewtype,
                                 *[
                                     x.literal_type
@@ -676,7 +676,9 @@ class type_getattr_record(numba.core.typing.templates.AttributeTemplate):
                                     else x
                                     for x in args
                                 ],
-                            )(lower)
+                            )(
+                                lower  # noqa: B023
+                            )
                             return sig
 
                 return numba.types.BoundFunction(type_method, recordviewtype)
