@@ -2,67 +2,16 @@
 
 # v2: keep this file, but modernize the 'of' function; ptr_lib is gone.
 
-from __future__ import absolute_import
-
 import ctypes
 
+from collections.abc import Iterable
+
 import numpy
 
 import awkward as ak
 
 
-def of(*arrays):
-    libs = set()
-    for array in arrays:
-        nplike = getattr(array, "nplike", None)
-        if isinstance(nplike, NumpyLike):
-            libs.add(nplike)
-        elif isinstance(array, numpy.ndarray):
-            ptr_lib = "cpu"
-        elif (
-            type(array).__module__.startswith("cupy.")
-            and type(array).__name__ == "ndarray"
-        ):
-            ptr_lib = "cuda"
-        else:
-            ptr_lib = ak.operations.convert.kernels(array)
-            if ptr_lib is None:
-                pass
-            elif ptr_lib == "cpu":
-                libs.add("cpu")
-            elif ptr_lib == "cuda":
-                libs.add("cuda")
-            else:
-                raise ValueError(
-                    """structure mixes 'cpu' and 'cuda' buffers; use one of
-
-    ak.to_kernels(array, 'cpu')
-    ak.to_kernels(array, 'cuda')
-
-to obtain an unmixed array in main memory or the GPU(s)."""
-                    + ak._util.exception_suffix(__file__)
-                )
-
-    if libs == set() or libs == set(["cpu"]):
-        return Numpy.instance()
-    elif libs == set(["cuda"]):
-        return Cupy.instance()
-    elif len(libs) == 1:
-        return next(iter(libs))
-    else:
-        raise ValueError(
-            """attempting to use both a 'cpu' array and a 'cuda' array in the """
-            """same operation; use one of
-
-    ak.to_kernels(array, 'cpu')
-    ak.to_kernels(array, 'cuda')
-
-to move one or the other to main memory or the GPU(s)."""
-            + ak._util.exception_suffix(__file__)
-        )
-
-
-class Singleton(object):
+class Singleton:
     _instance = None
 
     @classmethod
@@ -82,6 +31,7 @@ class NumpyMetadata(Singleton):
     uint16 = numpy.uint16
     uint32 = numpy.uint32
     uint64 = numpy.uint64
+    longlong = numpy.longlong
     float32 = numpy.float32
     float64 = numpy.float64
     complex64 = numpy.complex64
@@ -94,6 +44,7 @@ class NumpyMetadata(Singleton):
     signedinteger = numpy.signedinteger
     unsignedinteger = numpy.unsignedinteger
     floating = numpy.floating
+    complexfloating = numpy.complexfloating
     number = numpy.number
     object_ = numpy.object_
     generic = numpy.generic
@@ -131,12 +82,9 @@ if hasattr(numpy, "datetime64"):
 if hasattr(numpy, "timedelta64"):
     NumpyMetadata.timedelta64 = numpy.timedelta64
 
-NumpyMetadata.all_complex = tuple(
-    getattr(numpy, x) for x in dir(NumpyMetadata) if x.startswith("complex")
-)
-
 
 class NumpyLike(Singleton):
+    known_data = True
     known_shape = True
     known_dtype = True
 
@@ -227,10 +175,6 @@ class NumpyLike(Singleton):
         # array1[, array2[, ...]]
         return self._module.broadcast_arrays(*args, **kwargs)
 
-    def add(self, *args, **kwargs):
-        # array1, array2[, out=]
-        return self._module.add(*args, **kwargs)
-
     def cumsum(self, *args, **kwargs):
         # arrays[, out=]
         return self._module.cumsum(*args, **kwargs)
@@ -256,6 +200,10 @@ class NumpyLike(Singleton):
         # array1, array2
         return self._module.repeat(*args, **kwargs)
 
+    def tile(self, *args, **kwargs):
+        # array, int
+        return self._module.tile(*args, **kwargs)
+
     def stack(self, *args, **kwargs):
         # arrays
         return self._module.stack(*args, **kwargs)
@@ -290,6 +238,22 @@ class NumpyLike(Singleton):
 
     ############################ ufuncs
 
+    def add(self, *args, **kwargs):
+        # array1, array2
+        return self._module.add(*args, **kwargs)
+
+    def multiply(self, *args, **kwargs):
+        # array1, array2
+        return self._module.multiply(*args, **kwargs)
+
+    def logical_or(self, *args, **kwargs):
+        # array1, array2
+        return self._module.logical_or(*args, **kwargs)
+
+    def logical_and(self, *args, **kwargs):
+        # array1, array2
+        return self._module.logical_and(*args, **kwargs)
+
     def sqrt(self, *args, **kwargs):
         # array
         return self._module.sqrt(*args, **kwargs)
@@ -306,10 +270,6 @@ class NumpyLike(Singleton):
         # array1, array2[, out=output]
         return self._module.bitwise_or(*args, **kwargs)
 
-    def logical_and(self, *args, **kwargs):
-        # array1, array2
-        return self._module.logical_and(*args, **kwargs)
-
     def equal(self, *args, **kwargs):
         # array1, array2
         return self._module.equal(*args, **kwargs)
@@ -318,6 +278,14 @@ class NumpyLike(Singleton):
         # array
         return self._module.ceil(*args, **kwargs)
 
+    def minimum(self, *args, **kwargs):
+        # array1, array2
+        return self._module.minimum(*args, **kwargs)
+
+    def maximum(self, *args, **kwargs):
+        # array1, array2
+        return self._module.maximum(*args, **kwargs)
+
     ############################ almost-ufuncs
 
     def nan_to_num(self, *args, **kwargs):
@@ -328,10 +296,27 @@ class NumpyLike(Singleton):
         # a, b, rtol=1e-05, atol=1e-08, equal_nan=False
         return self._module.isclose(*args, **kwargs)
 
+    def isnan(self, *args, **kwargs):
+        # array
+        return self._module.isnan(*args, **kwargs)
+
+    def isneginf(self, *args, **kwargs):
+        # array
+        return self._module.isneginf(*args, **kwargs)
+
+    def isposinf(self, *args, **kwargs):
+        # array
+        return self._module.isposinf(*args, **kwargs)
+
+    def isfinite(self, *args, **kwargs):
+        # array
+        return self._module.isfinite(*args, **kwargs)
+
     ############################ reducers
 
     def all(self, *args, **kwargs):
         # array
+        kwargs.pop("prefer", None)
         return self._module.all(*args, **kwargs)
 
     def any(self, *args, **kwargs):
@@ -375,13 +360,13 @@ class NumpyLike(Singleton):
         return self._module.datetime_as_string(*args, **kwargs)
 
 
-class NumpyKernel(object):
+class NumpyKernel:
     def __init__(self, kernel, name_and_types):
         self._kernel = kernel
         self._name_and_types = name_and_types
 
     def __repr__(self):
-        return "<{0} {1}{2}>".format(
+        return "<{} {}{}>".format(
             type(self).__name__,
             self._name_and_types[0],
             "".join(", " + str(numpy.dtype(x)) for x in self._name_and_types[1:]),
@@ -390,8 +375,19 @@ class NumpyKernel(object):
     @staticmethod
     def _cast(x, t):
         if issubclass(t, ctypes._Pointer):
-            if isinstance(x, numpy.ndarray):
+
+            if is_numpy_buffer(x):
                 return ctypes.cast(x.ctypes.data, t)
+            elif is_cupy_buffer(x):
+                raise ak._v2._util.error(
+                    AssertionError("CuPy buffers shouldn't be passed to Numpy Kernels.")
+                )
+            elif is_jax_buffer(x):
+                raise ak._v2._util.error(
+                    ValueError(
+                        "JAX Buffers can't be passed as function args for the C Kernels"
+                    )
+                )
             else:
                 return ctypes.cast(x, t)
         else:
@@ -399,14 +395,95 @@ class NumpyKernel(object):
 
     def __call__(self, *args):
         assert len(args) == len(self._kernel.argtypes)
-        return self._kernel(
-            *(self._cast(x, t) for x, t in zip(args, self._kernel.argtypes))
+
+        if not any(is_jax_tracer(arg) for arg in args):
+            return self._kernel(
+                *(self._cast(x, t) for x, t in zip(args, self._kernel.argtypes))
+            )
+
+
+class CupyKernel(NumpyKernel):
+    def max_length(self, args):
+        cupy = ak._v2._connect.cuda.import_cupy("Awkward Arrays with CUDA")
+        max_length = numpy.iinfo(numpy.int64).min
+        for array in args:
+            if isinstance(array, cupy.ndarray):
+                max_length = max(max_length, len(array))
+        return max_length
+
+    def calc_grid(self, length):
+        if length > 1024:
+            return -(length // -1024), 1, 1
+        return 1, 1, 1
+
+    def calc_blocks(self, length):
+        if length > 1024:
+            return 1024, 1, 1
+        return length, 1, 1
+
+    def __call__(self, *args):
+        cupy = ak._v2._connect.cuda.import_cupy("Awkward Arrays with CUDA")
+        maxlength = self.max_length(args)
+        grid, blocks = self.calc_grid(maxlength), self.calc_blocks(maxlength)
+        cupy_stream_ptr = cupy.cuda.get_current_stream().ptr
+
+        if cupy_stream_ptr not in ak._v2._connect.cuda.cuda_streamptr_to_contexts:
+            ak._v2._connect.cuda.cuda_streamptr_to_contexts[cupy_stream_ptr] = (
+                cupy.array(ak._v2._connect.cuda.NO_ERROR),
+                [],
+            )
+
+        assert len(args) == len(self._kernel.dir)
+        # The first arg is the invocation index which raises itself by 8 in the kernel if there was no error before.
+        # The second arg is the error_code.
+        args = list(args)
+        args.extend(
+            [
+                len(
+                    ak._v2._connect.cuda.cuda_streamptr_to_contexts[cupy_stream_ptr][1]
+                ),
+                ak._v2._connect.cuda.cuda_streamptr_to_contexts[cupy_stream_ptr][0],
+            ]
         )
+        ak._v2._connect.cuda.cuda_streamptr_to_contexts[cupy_stream_ptr][1].append(
+            ak._v2._connect.cuda.Invocation(
+                name=self._name_and_types[0],
+                error_context=ak._v2._util.ErrorContext.primary(),
+            )
+        )
+
+        self._kernel(grid, blocks, tuple(args))
 
 
 class Numpy(NumpyLike):
+    @property
+    def index_nplike(self):
+        return self
+
     def to_rectilinear(self, array, *args, **kwargs):
-        return ak.operations.convert.to_numpy(array, *args, **kwargs)
+        if isinstance(array, numpy.ndarray):
+            return array
+
+        elif isinstance(
+            array,
+            (
+                ak.Array,
+                ak.Record,
+                ak.ArrayBuilder,
+                ak.layout.Content,
+                ak.layout.Record,
+                ak.layout.ArrayBuilder,
+                ak.layout.LayoutBuilder32,
+                ak.layout.LayoutBuilder64,
+            ),
+        ):
+            return ak.operations.convert.to_numpy(array, *args, **kwargs)
+
+        elif isinstance(array, Iterable):
+            return [self.to_rectilinear(x, *args, **kwargs) for x in array]
+
+        else:
+            raise TypeError("to_rectilinear argument must be iterable")
 
     def __getitem__(self, name_and_types):
         return NumpyKernel(ak._cpu_kernels.kernel[name_and_types], name_and_types)
@@ -426,28 +503,50 @@ class Numpy(NumpyLike):
     def ndarray(self):
         return self._module.ndarray
 
+    def raw(self, array, nplike):
+        if isinstance(nplike, Numpy):
+            return array
+        elif isinstance(nplike, Cupy):
+            cupy = Cupy.instance()
+            return cupy.asarray(array, dtype=array.dtype, order="C")
+        elif isinstance(nplike, ak._v2._typetracer.TypeTracer):
+            return ak._v2._typetracer.TypeTracerArray(
+                dtype=array.dtype, shape=array.shape
+            )
+        elif isinstance(nplike, Jax):
+            jax = Jax.instance()
+            return jax.asarray(array, dtype=array.dtype)
+        else:
+            raise TypeError(
+                "Invalid nplike, choose between nplike.Numpy, nplike.Cupy, Typetracer or Jax"
+            )
+
 
 class Cupy(NumpyLike):
+    @property
+    def index_nplike(self):
+        return self
+
     def to_rectilinear(self, array, *args, **kwargs):
         return ak.operations.convert.to_cupy(array, *args, **kwargs)
 
     def __getitem__(self, name_and_types):
-        raise NotImplementedError("no CUDA in v2 yet")
+        cupy = ak._v2._connect.cuda.import_cupy("Awkward Arrays with CUDA")
+        _cuda_kernels = ak._v2._connect.cuda.initialize_cuda_kernels(cupy)  # noqa: F401
+
+        func = _cuda_kernels[name_and_types]
+        if func is not None:
+            return CupyKernel(func, name_and_types)
+        else:
+            raise NotImplementedError(
+                f"{name_and_types[0]} is not implemented for CUDA. Please transfer the array back to the Main Memory to "
+                "continue the operation."
+            )
 
     def __init__(self):
-        try:
-            import cupy
-        except ImportError:
-            raise ImportError(
-                """to use CUDA arrays in Python, install the 'cupy' package with:
+        import awkward._v2._connect.cuda  # noqa: F401
 
-    pip install cupy --upgrade
-
-or
-
-    conda install cupy"""
-            )
-        self._module = cupy
+        self._module = ak._v2._connect.cuda.import_cupy("Awkward Arrays with CUDA")
 
     @property
     def ma(self):
@@ -467,7 +566,7 @@ or
     def ndarray(self):
         return self._module.ndarray
 
-    def asarray(self, array, dtype=None):
+    def asarray(self, array, dtype=None, order=None):
         if isinstance(
             array,
             (
@@ -479,11 +578,29 @@ or
         ):
             out = ak.operations.convert.to_cupy(array)
             if dtype is not None and out.dtype != dtype:
-                return self._module.asarray(out, dtype=dtype)
+                return self._module.asarray(out, dtype=dtype, order=order)
             else:
                 return out
         else:
-            return self._module.asarray(array, dtype=dtype)
+            return self._module.asarray(array, dtype=dtype, order=order)
+
+    def raw(self, array, nplike):
+        if isinstance(nplike, Cupy):
+            return array
+        elif isinstance(nplike, Numpy):
+            numpy = Numpy.instance()
+            return numpy.asarray(array.get(), dtype=array.dtype, order="C")
+        elif isinstance(nplike, ak._v2._typetracer.TypeTracer):
+            return ak._v2._typetracer.TypeTracerArray(
+                dtype=array.dtype, shape=array.shape
+            )
+        elif isinstance(nplike, Jax):
+            jax = Jax.instance()
+            return jax.asarray(array.get(), dtype=array.dtype)
+        else:
+            raise TypeError(
+                "Invalid nplike, choose between nplike.Numpy, nplike.Cupy, Typetracer or Jax"
+            )
 
     def ascontiguousarray(self, array, dtype=None):
         if isinstance(
@@ -529,25 +646,8 @@ or
         else:
             return self._module.repeat(array, repeats)
 
-    def nan_to_num(self, array, copy=True, nan=0.0, posinf=None, neginf=None):
-        # https://github.com/cupy/cupy/issues/4867
-        if copy:
-            array = self._module.copy(array)
-        if posinf is None:
-            if array.dtype.kind == "f":
-                posinf = numpy.finfo(array.dtype.type).max
-            else:
-                posinf = numpy.iinfo(array.dtype.type).max
-        if neginf is None:
-            if array.dtype.kind == "f":
-                neginf = numpy.finfo(array.dtype.type).min
-            else:
-                neginf = numpy.iinfo(array.dtype.type).min
-
-        array[self._module.isnan(array)] = nan
-        array[self._module.isinf(array) & (array > 0)] = posinf
-        array[self._module.isinf(array) & (array < 0)] = neginf
-        return array
+    def nan_to_num(self, *args, **kwargs):
+        self._module.nan_to_num(*args, **kwargs)
 
     # For all reducers: https://github.com/cupy/cupy/issues/3819
 
@@ -621,3 +721,203 @@ or
     ):
         # array, max_line_width, precision=None, suppress_small=None
         return self._module.array_str(array, max_line_width, precision, suppress_small)
+
+
+class Jax(NumpyLike):
+    @property
+    def index_nplike(self):
+        return ak.nplike.Numpy.instance()
+
+    def to_rectilinear(self, array, *args, **kwargs):
+        if isinstance(array, self._module.DeviceArray):
+            return array
+
+        elif isinstance(
+            array,
+            (
+                ak.Array,
+                ak.Record,
+                ak.ArrayBuilder,
+                ak.layout.Content,
+                ak.layout.Record,
+                ak.layout.ArrayBuilder,
+                ak.layout.LayoutBuilder32,
+                ak.layout.LayoutBuilder64,
+            ),
+        ):
+            return ak.operations.convert.to_jax(array, *args, **kwargs)
+
+        elif isinstance(array, Iterable):
+            return [self.to_rectilinear(x, *args, **kwargs) for x in array]
+
+        else:
+            raise ak._v2._util.error(
+                ValueError("to_rectilinear argument must be iterable")
+            )
+
+    def __getitem__(self, name_and_types):
+        return NumpyKernel(ak._cpu_kernels.kernel[name_and_types], name_and_types)
+
+    def __init__(self):
+        from awkward._v2._connect.jax import import_jax  # noqa: F401
+
+        self._module = import_jax().numpy
+
+    @property
+    def ma(self):
+        ak._v2._util.error(
+            ValueError(
+                "JAX arrays cannot have missing values until JAX implements "
+                "numpy.ma.MaskedArray" + ak._util.exception_suffix(__file__)
+            )
+        )
+
+    @property
+    def char(self):
+        ak._v2._util.error(
+            ValueError(
+                "JAX arrays cannot do string manipulations until JAX implements "
+                "numpy.char"
+            )
+        )
+
+    @property
+    def ndarray(self):
+        return self._module.ndarray
+
+    def asarray(self, array, dtype=None, order=None):
+        return self._module.asarray(array, dtype=dtype, order="K")
+
+    def ascontiguousarray(self, array, dtype=None):
+        if isinstance(
+            array,
+            (
+                ak.highlevel.Array,
+                ak.highlevel.Record,
+                ak.layout.Content,
+                ak.layout.Record,
+            ),
+        ):
+            out = ak.operations.convert.to_jax(array)
+            if dtype is not None and out.dtype != dtype:
+                return self._module.ascontiguousarray(out, dtype=dtype)
+            else:
+                return out
+        else:
+            return self._module.ascontiguousarray(array, dtype=dtype)
+
+    def raw(self, array, nplike):
+        if isinstance(nplike, Jax):
+            return array
+        elif isinstance(nplike, ak.nplike.Cupy):
+            cupy = ak.nplike.Cupy.instance()
+            return cupy.asarray(array)
+        elif isinstance(nplike, ak.nplike.Numpy):
+            numpy = ak.nplike.Numpy.instance()
+            return numpy.asarray(array)
+        elif isinstance(nplike, ak._v2._typetracer.TypeTracer):
+            return ak._v2._typetracer.TypeTracerArray(
+                dtype=array.dtype, shape=array.shape
+            )
+        else:
+            ak._v2._util.error(
+                TypeError(
+                    "Invalid nplike, choose between nplike.Numpy, nplike.Cupy, Typetracer or Jax",
+                )
+            )
+
+    # For all reducers: JAX returns zero-dimensional arrays like CuPy
+
+    def all(self, *args, **kwargs):
+        out = self._module.all(*args, **kwargs)
+        return out
+
+    def any(self, *args, **kwargs):
+        out = self._module.any(*args, **kwargs)
+        return out
+
+    def count_nonzero(self, *args, **kwargs):
+        out = self._module.count_nonzero(*args, **kwargs)
+        return out
+
+    def sum(self, *args, **kwargs):
+        out = self._module.sum(*args, **kwargs)
+        return out
+
+    def prod(self, *args, **kwargs):
+        out = self._module.prod(*args, **kwargs)
+        return out
+
+    def min(self, *args, **kwargs):
+        out = self._module.min(*args, **kwargs)
+        return out
+
+    def max(self, *args, **kwargs):
+        out = self._module.max(*args, **kwargs)
+        return out
+
+    def argmin(self, *args, **kwargs):
+        out = self._module.argmin(*args, **kwargs)
+        return out
+
+    def argmax(self, *args, **kwargs):
+        out = self._module.argmax(*args, **kwargs)
+        return out
+
+
+def is_numpy_buffer(array):
+    return isinstance(array, numpy.ndarray)
+
+
+def is_cupy_buffer(array):
+    return type(array).__module__.startswith("cupy.")
+
+
+def is_jax_buffer(array):
+    return type(array).__module__.startswith("jaxlib.")
+
+
+def is_jax_tracer(tracer):
+    return type(tracer).__module__.startswith("jax.")
+
+
+def of(*arrays, default_cls=Numpy):
+    """
+    Args:
+        *arrays: iterable of possible array objects
+        default_cls: default NumpyLike class if no array objects found
+
+    Return the #ak.nplike.NumpyLike that is best-suited to operating upon the given
+    iterable of arrays. Return an instance of the `default_cls` if no known array types
+    are found.
+    """
+    nplikes = set()
+    for array in arrays:
+        nplike = getattr(array, "nplike", None)
+        if nplike is not None:
+            nplikes.add(nplike)
+        elif is_numpy_buffer(array):
+            nplikes.add(Numpy.instance())
+        elif is_cupy_buffer(array):
+            nplikes.add(Cupy.instance())
+        elif is_jax_buffer(array):
+            nplikes.add(Jax.instance())
+
+    if any(isinstance(x, ak._v2._typetracer.TypeTracer) for x in nplikes):
+        return ak._v2._typetracer.TypeTracer.instance()
+
+    if nplikes == set():
+        return default_cls.instance()
+    elif len(nplikes) == 1:
+        return next(iter(nplikes))
+    else:
+        raise ValueError(
+            """attempting to use both a 'cpu' array and a 'cuda' array in the """
+            """same operation; use one of
+
+    ak.to_backend(array, 'cpu')
+    ak.to_backend(array, 'cuda')
+
+to move one or the other to main memory or the GPU(s)."""
+            + ak._util.exception_suffix(__file__)
+        )
